import subprocess
import os
import git

__version__ = '1.5.1'
name = 'statespace'
image = 'noahhsmith/' + name + ':' + __version__
dist = 'dist/' + name + '-' + __version__ + '.tar.gz'
pathdockerpwd = '../../.dockerpwd'


def main():
    repo = git.Repo('.', search_parent_directories=True)
    os.chdir(repo.working_tree_dir)
    # dockerhub()
    pypi()


def runner(cmd):
    return subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


def dockerhub():
    res1 = runner(['docker', 'build', '-t', image, '.'])
    with open(pathdockerpwd, 'rt') as fh: pwd = fh.read().strip()
    res2 = runner(['docker', 'login', '-u', 'noahhsmith', '-p', pwd])
    res3 = runner(['docker', 'push', image])
    return


def pypi():
    res1 = runner(['python3', 'setup.py', 'build_ext'])
    res2 = runner(['python3', 'setup.py', 'build_py'])
    res3 = runner(['python3', 'setup.py', 'sdist'])
    res4 = runner(['python3', '-m', 'twine', 'upload', '--repository', 'pypi', dist])
    return


if __name__ == '__main__':
    main()



from statespace.processors.kalman import Kalman
from statespace.processors.sigmapoint import SigmaPoint
from statespace.processors.particle import Particle
from statespace.models.onestate import Onestate
from statespace.models.threestate import Threestate
from statespace.models.bearingsonly import BearingsOnly

cases = {
    '1a': {'model': Onestate(), 'processor': Kalman()},
    '1b': {'model': Threestate(), 'processor': Kalman()},
    '1c': {'model': Onestate(), 'processor': Kalman('ud')},
    '1d': {'model': Threestate(), 'processor': Kalman('ud')},
    '2a': {'model': Onestate(), 'processor': SigmaPoint()},
    '2b': {'model': Threestate(), 'processor': SigmaPoint()},
    '2c': {'model': Onestate(), 'processor': SigmaPoint('cholesky')},
    '2d': {'model': Threestate(), 'processor': SigmaPoint('cholesky')},
    '3a': {'model': Onestate(), 'processor': Particle()},
    '3b': {'model': Threestate(), 'processor': Particle()},
    'bo1': {'model': BearingsOnly(), 'processor': Kalman()},
    'bo2': {'model': BearingsOnly(), 'processor': SigmaPoint()},  # tuning
    'bo3': {'model': BearingsOnly(), 'processor': Particle()},  # tuning
}
case = '1b'


def main():
    model, processor = cases[case]['model'], cases[case]['processor']
    processor.run(model)
    model.eval.estimate(processor.log)
    # model.eval.autocorr.run(processor.log)
    model.eval.show()


if __name__ == "__main__":
    main()



import math
import numpy as np
from scipy.linalg.blas import drot, drotg


class SigmaPoint:
    '''modern sigma-point deterministic sampling kalman filter'''

    def __init__(self, *args, **kwargs):
        self.args, self.kwargs, self.log = args, kwargs, []
        if 'cholesky' in args:
            self.run = self.spfcholesky
        else:
            self.run = self.spf

    def spf(self, model):
        '''sigma-point deterministic sampling kalman filter'''
        sim, f, h, F, H, R, Q, G, x, P = model.ekf()
        XY, W, WM = model.sp()
        for t, o, u in sim():
            X, Y = XY(x, P, u)
            x, y = np.sum(np.multiply(WM, X), axis=1).reshape(-1, 1), np.sum(np.multiply(W, Y), axis=1)[0]
            Xres, Yres = self.Xres(X, x), self.Yres(Y, y)
            P = np.multiply(WM, Xres) @ Xres.T + G @ Q @ G.T
            K = np.multiply(WM, Xres) @ Yres.T / (np.multiply(W, Yres) @ Yres.T + R)
            x = x + K * (o - y)
            P = P - K @ (np.multiply(W, Yres) @ Yres.T + R) @ K.T
            self.log.append([t, x, y])

    def spfcholesky(self, model):
        '''cholesky factorized sigma-point sampling kalman filter'''
        sim, f, h, F, H, R, Q, G, x, P = model.ekf()
        XY, W, X2, Y2, P2, S, Sp, So = model.spcho()
        for t, o, u in sim():
            x, S, X, Y = self.temporal(x, XY, W, X2, S, Sp, u)
            x, S, y = self.observational(x, o, X, Y, W, X2, Y2, P2, S, So)
            self.log.append([t, x, y])

    def temporal(self, x, XY, W, Xtil, S, Sproc, u):
        '''cholesky factorized temporal update'''
        X, Y = XY(x, S, u)
        x = X @ W.T
        for i in range(X.shape[1]): Xtil[:, i] = (X[:, i].reshape(-1, 1) - x).T
        q, r = np.linalg.qr(np.concatenate([math.sqrt(W[0, 1]) * Xtil[:, 1:], Sproc], 1))
        S = self.cholupdate(r.T[0:X.shape[0], 0:X.shape[0]], W[0, 0] * Xtil[:, 0].reshape(-1, 1))
        return x, S, X, Y

    def observational(self, x, o, X, Y, W, Xtil, Ytil, Pxy, S, Sobs):
        '''cholesky factorized observational update'''
        y = (W @ Y.T)[0, 0]
        for i in range(X.shape[1]): Ytil[0, i] = (Y[0, i].reshape(-1, 1) - y).T
        q, r = np.linalg.qr(np.concatenate([math.sqrt(W[0, 1]) * Ytil[:, 1:], Sobs], 1))
        Sy = self.cholupdate(r.T[0:1, 0:1], W[0] * Ytil[:, 0].reshape(-1, 1))
        for i in range(X.shape[1]):
            tmp = W[0, i] * Xtil[:, i].reshape(-1, 1) * Ytil[0, i]
            Pxy += tmp
        if Sy[0, 0] < math.sqrt(10) or Sy[0, 0] > math.sqrt(1000): Sy[0, 0] = math.sqrt(1000)
        K = Pxy / Sy[0, 0] ** 2
        U = K * Sy
        x = x + K * (o - y)
        S = self.choldowndate(S, U)
        return x, S, y

    def cholupdate(self, R, z):
        '''cholesky update'''
        n = z.shape[0]
        for k in range(n):
            c, s = drotg(R[k, k], z[k])
            drot(R[k, :], z, c, s, overwrite_x=True, overwrite_y=True)
        return R

    def choldowndate(self, R, z):
        '''cholesky downdate'''
        n = R.shape[0]
        for k in range(n):
            if (R[k, k] - z[k]) * (R[k, k] + z[k]) < 0: return R
            rbar = np.sqrt((R[k, k] - z[k]) * (R[k, k] + z[k]))
            for j in range(k + 1, n):
                R[k, j] = 1. / rbar * (R[k, k] * R[k, j] - z[k] * z[j])
                z[j] = 1. / R[k, k] * (rbar * z[j] - z[k] * R[k, j])
            R[k, k] = rbar
        return R

    def Xres(self, X, x):
        X2 = X.copy()
        for i in range(X2.shape[1]):
            for j in range(X2.shape[0]): X2[j, i] -= x[j, 0]
        return X2

    def Yres(self, Y, y):
        for i in range(Y.shape[1]): Y[0, i] -= y
        return Y


if __name__ == "__main__":
    pass



from statespace.models.threestate import Threestate
from statespace.processors.kalman import Kalman


def test_basic():
    model, processor = Threestate(), Kalman()
    result = processor.run(model)
    assert result



import os
import sys
import subprocess
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from statespace.version import __version__


def main():
    with open('requirements.txt') as fin: required = fin.read().splitlines()
    with open('README.md', 'r') as fh: long_description = fh.read()
    # noinspection PyTypeChecker
    setup(
        name='statespace',
        version=__version__,
        author='noah smith',
        author_email='noahhsmith@gmail.com',
        description='statespace processors and models',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://gitlab.com/noahhsmith/statespace',
        packages=find_packages(),
        entry_points={'console_scripts': ['statespace = statespace.__main__:main']},
        ext_modules=[CMakeExtension('statespace')],
        cmdclass=dict(build_ext=CMakeBuild),
        install_requires=required,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent", ],
    )

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))
        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir, '-DPYTHON_EXECUTABLE=' + sys.executable]
        cfg = 'Release'  # 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]
        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
        build_args += ['--', '-j2']
        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''), __version__)
        if not os.path.exists(self.build_temp): os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)
        print()


if __name__ == '__main__':
    main()
